/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __houdini_SurfaceAccessorHoudini_h__
#define __houdini_SurfaceAccessorHoudini_h__

namespace fe
{

// TODO rework select() methodology to be reentrant

/**************************************************************************//**
    @brief Houdini Surface Editing Implementation

	Note that attempts to access non-existant attributes isn't a failure.
	New attributes may be added when first written to by a set() method.

	@ingroup houdini
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorHoudini : virtual public SurfaceAccessorI
{
	public:
						SurfaceAccessorHoudini(void):
							m_pGdp(NULL)									{}

virtual					~SurfaceAccessorHoudini(void)						{}

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							//* see standard attributes in GEO/GEO_Detail.h

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="P";
									break;
								case SurfaceAccessibleI::e_normal:
									name=GEO_STD_ATTRIB_NORMAL;
									break;
								case SurfaceAccessibleI::e_vertices:
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									m_element=a_element;
									return TRUE;
							}
							bindInternal(a_element,name);
							return TRUE;	//* NOTE allow unbound attributes
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_attribute=SurfaceAccessibleI::e_generic;
							bindInternal(a_element,a_name);
							return TRUE;	//* NOTE allow unbound attributes
						}

						//* as SurfaceAccessorI
virtual	U32				count(void) const
						{
							FEASSERT(m_pGdp);

							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
								return m_pGdp->points().entries();

//							case SurfaceAccessibleI::e_vertex:
//								return 0;

							case SurfaceAccessibleI::e_primitive:
								return m_pGdp->primitives().entries();

//							case SurfaceAccessibleI::e_detail:
//								return 0;

							default:
								;
							}
							return 0;
						}
virtual	U32				subCount(U32 index) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							const_cast<SurfaceAccessorHoudini*>(this)->select(
									index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								return m_pPrimitive->getVertexCount();
							}

							return m_handle.entries();
						}

virtual	void			set(U32 index,I32 a_integer)
						{
							set(index,0,a_integer);
						}
virtual	void			set(U32 index,U32 subIndex,I32 a_integer)
						{
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addInteger();
							}
							select(index);

							m_handle.setI(a_integer,subIndex);
						}
virtual	I32				integer(U32 index,U32 subIndex=0)
						{
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								return m_pPrimitive->getVertex(subIndex)
										.getBasePt()->getNum();
							}

							return m_handle.getI(subIndex);
						}

virtual	void			set(U32 index,Real a_real)
						{
							set(index,0,a_real);
						}
virtual	void			set(U32 index,U32 subIndex,Real a_real)
						{
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addReal();
							}
							select(index);

							m_handle.setF(a_real,subIndex);
						}
virtual	Real			real(U32 index,U32 subIndex=0)
						{
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								return 0.0;
							}

							return m_handle.getF(subIndex);
						}

virtual	void			set(U32 index,const SpatialVector& a_vector)
						{
							set(index,0,a_vector);
						}
virtual	void			set(U32 index,U32 subIndex,
								const SpatialVector& a_vector)
						{
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addSpatialVector();
							}
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								UT_Vector4 utVector4(a_vector.temp());
								m_pPrimitive->getVertex(subIndex).getPt()
										->setPos(utVector4);
								return;
							}

							UT_Vector4 utVector4(a_vector.temp());
							m_handle.setV4(utVector4);
						}
virtual	SpatialVector	spatialVector(U32 index,U32 subIndex=0)
						{
							select(index);

							// GEO_Vertex* pVertex=prim->getVertex(subIndex);
							// GEO_Point* pPoint=pVertex->getPt();
							// GB_Element* pElement=pVertex->getBasePt();

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								UT_Vector4 utVector4=m_pPrimitive
										->getVertex(subIndex).getPos();
								SpatialVector vector(utVector4.data());
								return vector;
							}

							UT_Vector4 utVector4=m_handle.getV4(subIndex);
							SpatialVector vector(utVector4.data());
							return vector;
						}

						//* Houdini specific
		void			setGdp(const GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=NULL;
						}
		void			setGdpChangeable(GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=a_pGdp;
						}
const	GU_Detail*		gdp(void)				{ return m_pGdp; }
		GU_Detail*		gdpChangeable(void)		{ return m_pGdpChangeable; }

	private:
		void			addInteger(void);
		void			addReal(void);
		void			addSpatialVector(void);

		BWORD			isBound(void) const
						{	return m_handle.isAttributeValid(); }
		BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							FEASSERT(m_pGdp);
							m_element=SurfaceAccessibleI::e_point;
							if(a_element>=0 && a_element<=4)
							{
								m_element=a_element;
							}
							GEO_AttributeOwner owner=ms_elementTable[m_element];
							m_name=a_name;
							m_handle=
									m_pGdp->getAttribute(owner,a_name.c_str());
							return isBound();
						}

		void			select(U32 index)
						{
							FEASSERT(m_pGdp);
							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
								m_handle.setElement(m_pGdp->points()(index));
								break;

//							case SurfaceAccessibleI::e_vertex:
//								break;

							case SurfaceAccessibleI::e_primitive:
								m_pPrimitive=m_pGdp->primitives()(index);
								m_handle.setElement(m_pPrimitive);
								break;

//							case SurfaceAccessibleI::e_detail:
//								break;

							default:
								;
							}
						}

		SurfaceAccessibleI::Element		m_element;
		SurfaceAccessibleI::Attribute	m_attribute;
		String							m_name;	//* in case we need to add it

const	GU_Detail*						m_pGdp;
		GU_Detail*						m_pGdpChangeable;

const	GEO_Primitive*					m_pPrimitive;
		GEO_AttributeHandle				m_handle;

static
const	GEO_AttributeOwner	ms_elementTable[];
};

} // namespace

#endif /* __houdini_SurfaceAccessorHoudini_h__ */
